@@ -1,6 +1,6 @@
 // info.h
 
-// Copyright (C) 2012-2016 by Werner Lemberg.
+// Copyright (C) 2012-2017 by Werner Lemberg.
 //
 // This file is part of the ttfautohint library, and may only be used,
 // modified, and distributed under the terms given in `COPYING'.  By
@@ -49,11 +49,14 @@ typedef struct Info_Data_
   bool hint_composites;
   char default_script[5];
   char fallback_script[5];
+  bool fallback_scaling;
   bool symbol;
   bool dehint;
   bool TTFA_info;
 
   const char* control_name;
+  const char* reference_name;
+  int reference_index;
 } Info_Data;
 
 
