@@ -1,6 +1,6 @@
 // main.cpp
 
-// Copyright (C) 2011-2016 by Werner Lemberg.
+// Copyright (C) 2011-2017 by Werner Lemberg.
 //
 // This file is part of the ttfautohint library, and may only be used,
 // modified, and distributed under the terms given in `COPYING'.  By
@@ -35,10 +35,13 @@
 #include <vector>
 #include <string>
 
-#if BUILD_GUI
+#ifdef BUILD_GUI
 #  include <QApplication>
 #  include "maingui.h"
 #else
+#  include <ft2build.h>
+#  include FT_FREETYPE_H
+#  include FT_TRUETYPE_TABLES_H // for option `-T'
 #  include "info.h"
 #endif
 
@@ -46,16 +49,18 @@
 #include <numberset.h>
 
 
-#ifdef _WIN32
-#  include <fcntl.h>
-#  define SET_BINARY(f) do { \
-                          if (!isatty(fileno(f))) \
-                            setmode(fileno(f), O_BINARY); \
-                        } while (0)
-#endif
+#ifndef BUILD_GUI
+#  ifdef _WIN32
+#    include <fcntl.h>
+#    define SET_BINARY(f) do { \
+                            if (!isatty(fileno(f))) \
+                              setmode(fileno(f), O_BINARY); \
+                          } while (0)
+#  endif
 
-#ifndef SET_BINARY
-#  define SET_BINARY(f) do {} while (0)
+#  ifndef SET_BINARY
+#    define SET_BINARY(f) do {} while (0)
+#  endif
 #endif
 
 
@@ -82,17 +87,19 @@ const Tag_Names script_names[] =
 };
 
 
-// the available feature tags and its descriptions are directly extracted
-// from `ttfautohint-coverages.h'
-#undef COVERAGE
-#define COVERAGE(n, N, d, t, t1, t2, t3, t4) \
-          {#t, d},
+#ifndef BUILD_GUI
+  // the available feature tags and its descriptions are directly extracted
+  // from `ttfautohint-coverages.h'
+#  undef COVERAGE
+#  define COVERAGE(n, N, d, t, t1, t2, t3, t4) \
+            {#t, d},
 
 const Tag_Names feature_names[] =
 {
 #include <ttfautohint-coverages.h>
   {NULL, NULL}
 };
+#endif
 
 
 #ifndef BUILD_GUI
@@ -131,7 +138,7 @@ progress(long curr_idx,
   }
 
   // print progress approx. every 10%
-  int curr_percent = curr_idx * 100 / num_glyphs;
+  int curr_percent = int(curr_idx * 100 / num_glyphs);
   int curr_diff = curr_percent - data->last_percent;
 
   if (curr_diff >= 10)
@@ -202,9 +209,13 @@ err(TA_Error error,
   else
   {
     if (error < 0x100)
+    {
       fprintf(stderr, "An error with code 0x%02x occurred"
                         " while autohinting fonts\n",
                       error);
+      if (error_string)
+        fprintf(stderr, " %s", error_string);
+    }
     else if (error >= 0x100 && error < 0x200)
     {
       fprintf(stderr, "An error with code 0x%03x occurred"
@@ -232,6 +243,15 @@ err(TA_Error error,
       if (errpos && errline)
         fprintf(stderr, "  %*s\n", int(errpos - errline + 1), "^");
     }
+    else if (error >= 0x300 && error < 0x400)
+    {
+      error -= 0x300;
+      fprintf(stderr, "An error with code 0x%02x occurred"
+                        " while loading the reference font\n",
+                      error);
+      if (error_string)
+        fprintf(stderr, " %s", error_string);
+    }
   }
 }
 
@@ -315,8 +335,15 @@ show_help(bool
   fprintf(handle,
 "  -r, --hinting-range-max=N  the maximum PPEM value for hint sets\n"
 "                             (default: %d)\n"
+#ifndef BUILD_GUI
+"  -R, --reference=FILE       derive blue zones from reference font FILE\n"
+#endif
 "  -s, --symbol               input is symbol font\n"
+"  -S, --fallback-scaling     use fallback scaling, not hinting\n"
 "  -t, --ttfa-table           add TTFA information table\n"
+#ifndef BUILD_GUI
+"  -T, --ttfa-info            display TTFA table in IN-FILE and exit\n"
+#endif
 "  -v, --verbose              show progress information\n"
 "  -V, --version              print version information and exit\n"
 "  -w, --strong-stem-width=S  use strong stem width routine for modes S,\n"
@@ -334,6 +361,9 @@ show_help(bool
 "                             specify a comma-separated list of\n"
 "                             x-height snapping exceptions, for example\n"
 "                             \"-9, 13-17, 19\" (default: \"\")\n"
+#ifndef BUILD_GUI
+"  -Z, --reference-index=N    face index of reference font (default: 0)\n"
+#endif
 "\n",
           TA_HINTING_RANGE_MAX, TA_INCREASE_X_HEIGHT);
 
@@ -407,9 +437,12 @@ show_help(bool
 "\n");
   fprintf(handle,
 "Options -f and -D take a four-letter string that identifies a script.\n"
-"Option -f sets the script used as a fallback for glyphs that have\n"
-"character codes outside of known script ranges.  Option -D sets the\n"
-"default script for handling OpenType features.  Possible values are\n"
+"Option -f sets the script used as a fallback for glyphs that can't be\n"
+"associated with a known script.  By default, such glyphs are hinted;\n"
+"if option -S is set, they are scaled only instead.  Option -D sets the\n"
+"default script for handling OpenType features.\n"
+"\n"
+"Possible four-letter string values are\n"
 "\n");
   const Tag_Names* sn = script_names;
   for(;;)
@@ -501,7 +534,7 @@ show_version()
 #else
 "ttfautohint " VERSION "\n"
 #endif
-"Copyright (C) 2011-2016 Werner Lemberg <wl@gnu.org>.\n"
+"Copyright (C) 2011-2017 Werner Lemberg <wl@gnu.org>.\n"
 "License: FreeType License (FTL) or GNU GPLv2.\n"
 "This is free software: you are free to change and redistribute it.\n"
 "There is NO WARRANTY, to the extent permitted by law.\n");
@@ -511,6 +544,149 @@ show_version()
 #endif // CONSOLE_OUTPUT
 
 
+#ifndef BUILD_GUI
+
+typedef const struct FT_error_
+{
+  int err_code;
+  const char* err_msg;
+} FT_error;
+
+static FT_error FT_errors[] =
+
+#undef __FTERRORS_H__
+#define FT_ERRORDEF(e, v, s) { e, s },
+#define FT_ERROR_START_LIST {
+#define FT_ERROR_END_LIST { 0, NULL } };
+#include FT_ERRORS_H
+
+
+static const char*
+FT_get_error_message(FT_Error error)
+{
+  FT_error* e = FT_errors;
+
+  while (e->err_code || e->err_msg)
+  {
+    if (e->err_code == error)
+      return e->err_msg;
+    e++;
+  }
+
+  return NULL;
+}
+
+
+#define BUF_SIZE 0x10000
+
+#define TTAG_TTFA FT_MAKE_TAG('T', 'T', 'F', 'A')
+
+static void
+display_TTFA(FILE* in)
+{
+  FT_Byte buf[BUF_SIZE];
+  FT_Byte* in_buf;
+  size_t in_len = 0;
+  size_t read_bytes;
+
+  if (in == stdin)
+    SET_BINARY(stdin);
+
+  in_buf = (FT_Byte*)malloc(BUF_SIZE);
+  if (!in_buf)
+  {
+    fprintf(stderr, "Can't allocate enough memory.\n");
+    exit(EXIT_FAILURE);
+  }
+
+  while ((read_bytes = fread(buf, 1, BUF_SIZE, in)) > 0)
+  {
+    FT_Byte* in_buf_new = (FT_Byte*)realloc(in_buf, in_len + read_bytes);
+    if (!in_buf_new)
+    {
+      fprintf(stderr, "Can't reallocate enough memory.\n");
+      exit(EXIT_FAILURE);
+    }
+    else
+      in_buf = in_buf_new;
+
+    memcpy(in_buf + in_len, buf, read_bytes);
+
+    in_len += read_bytes;
+  }
+
+  if (ferror(in))
+  {
+    fprintf(stderr, "Stream error while handling input font.\n");
+    exit(EXIT_FAILURE);
+  }
+
+  FT_Library library = NULL;
+  FT_Face face = NULL;
+  FT_Error error;
+
+  error = FT_Init_FreeType(&library);
+  if (error)
+  {
+    fprintf(stderr, "Can't initialize FreeType library:\n"
+                    "%s\n",
+                    FT_get_error_message(error));
+    exit(EXIT_FAILURE);
+  }
+
+  // in a TTC, a `TTFA' table is part of the first subfont,
+  // thus we can simply pass 0 as the face index
+  error = FT_New_Memory_Face(library, in_buf, (FT_Long)in_len, 0, &face);
+  if (error)
+  {
+    fprintf(stderr, "Can't open input font:\n"
+                    "%s\n",
+                    FT_get_error_message(error));
+    exit(EXIT_FAILURE);
+  }
+
+  FT_Byte* ttfa_buf = NULL;
+  FT_ULong ttfa_len = 0;
+
+  error = FT_Load_Sfnt_Table(face, TTAG_TTFA, 0, NULL, &ttfa_len);
+  if (error)
+  {
+    fprintf(stderr, "No `TTFA' table in font.\n");
+    goto Exit;
+  }
+
+  ttfa_buf = (FT_Byte*)malloc(ttfa_len);
+  if (!ttfa_buf)
+  {
+    fprintf(stderr, "Can't allocate enough memory.\n");
+    exit(EXIT_FAILURE);
+  }
+
+  error = FT_Load_Sfnt_Table(face, TTAG_TTFA, 0, ttfa_buf, &ttfa_len);
+  if (error)
+  {
+    fprintf(stderr, "Error loading `TTFA' table:\n"
+                    "%s\n",
+                    FT_get_error_message(error));
+    exit(EXIT_FAILURE);
+  }
+
+  fprintf(stdout, "%s", ttfa_buf);
+
+Exit:
+  FT_Done_Face(face);
+  FT_Done_FreeType(library);
+
+  free(in_buf);
+  free(ttfa_buf);
+  if (in != stdin)
+    fclose(in);
+
+  exit(EXIT_SUCCESS);
+}
+#endif
+
+
 int
 main(int argc,
      char** argv)
@@ -538,7 +714,11 @@ main(int argc,
   bool no_info = false;
   bool detailed_info = false;
   bool TTFA_info = false;
+#ifndef BUILD_GUI
+  bool show_TTFA_info = false;
+#endif
   bool symbol = false;
+  bool fallback_scaling = false;
 
   const char* default_script = NULL;
   bool have_default_script = false;
@@ -560,6 +740,10 @@ main(int argc,
   TA_Info_Post_Func info_post_func = info_post;
 
   const char* control_name = NULL;
+  const char* reference_name = NULL;
+  int reference_index = 0;
+
+  unsigned long long epoch = ULLONG_MAX;
 #endif
 
   // For real numbers (both parsing and displaying) we only use `.' as the
@@ -602,6 +786,7 @@ main(int argc,
       {"default-script", required_argument, NULL, 'D'},
       {"dehint", no_argument, NULL, 'd'},
       {"detailed-info", no_argument, NULL, 'I'},
+      {"fallback-scaling", no_argument, NULL, 'S'},
       {"fallback-script", required_argument, NULL, 'f'},
       {"fallback-stem-width", required_argument, NULL, 'H'},
       {"family-suffix", required_argument, NULL, 'F'},
@@ -612,9 +797,16 @@ main(int argc,
       {"increase-x-height", required_argument, NULL, 'x'},
       {"no-info", no_argument, NULL, 'n'},
       {"pre-hinting", no_argument, NULL, 'p'},
+#ifndef BUILD_GUI
+      {"reference", required_argument, NULL, 'R'},
+      {"reference-index", required_argument, NULL, 'Z'},
+#endif
       {"strong-stem-width", required_argument, NULL, 'w'},
       {"symbol", no_argument, NULL, 's'},
       {"ttfa-table", no_argument, NULL, 't'},
+#ifndef BUILD_GUI
+      {"ttfa-info", no_argument, NULL, 'T'},
+#endif
       {"verbose", no_argument, NULL, 'v'},
       {"version", no_argument, NULL, 'V'},
       {"windows-compatibility", no_argument, NULL, 'W'},
@@ -652,9 +844,9 @@ main(int argc,
     int option_index;
     int c = getopt_long_only(argc, argv,
 #ifdef BUILD_GUI
-                             "cdD:f:F:G:hH:iIl:npr:stVvw:Wx:X:",
+                             "cdD:f:F:G:hH:iIl:npr:sStvVw:Wx:X:",
 #else
-                             "cdD:f:F:G:hH:iIl:m:npr:stVvw:Wx:X:",
+                             "cdD:f:F:G:hH:iIl:m:npr:R:sStTvVw:Wx:X:Z:",
 #endif
                              long_options, &option_index);
     if (c == -1)
@@ -735,14 +927,30 @@ main(int argc,
       have_hinting_range_max = true;
       break;
 
+#ifndef BUILD_GUI
+    case 'R':
+      reference_name = optarg;
+      break;
+#endif
+
     case 's':
       symbol = true;
       break;
 
+    case 'S':
+      fallback_scaling = true;
+      break;
+
     case 't':
       TTFA_info = true;
       break;
 
+#ifndef BUILD_GUI
+    case 'T':
+      show_TTFA_info = true;
+      break;
+#endif
+
     case 'v':
 #ifndef BUILD_GUI
       progress_func = progress;
@@ -775,6 +983,12 @@ main(int argc,
       have_x_height_snapping_exceptions_string = true;
       break;
 
+#ifndef BUILD_GUI
+    case 'Z':
+      reference_index = atoi(optarg);
+      break;
+#endif
+
 #ifndef BUILD_GUI
     case DEBUG_OPTION:
       debug = true;
@@ -813,13 +1027,30 @@ main(int argc,
     have_default_script = false;
     have_fallback_script = false;
     have_fallback_stem_width = false;
+    have_hinting_limit = false;
     have_hinting_range_max = false;
     have_hinting_range_min = false;
-    have_hinting_limit = false;
     have_increase_x_height = false;
     have_x_height_snapping_exceptions_string = false;
   }
 
+#ifndef BUILD_GUI
+  if (show_TTFA_info)
+  {
+    // -T makes ttfautohint ignore even more options
+    have_default_script = false;
+    have_fallback_script = false;
+    have_fallback_stem_width = false;
+    have_family_suffix = false;
+    have_hinting_limit = false;
+    have_hinting_range_max = false;
+    have_hinting_range_min = false;
+    have_increase_x_height = false;
+    have_x_height_snapping_exceptions_string = false;
+    debug = false;
+  }
+#endif
+
   if (!have_default_script)
     default_script = "latn";
   if (!have_fallback_script)
@@ -835,12 +1066,58 @@ main(int argc,
   if (!have_x_height_snapping_exceptions_string)
     x_height_snapping_exceptions_string = "";
   if (!have_fallback_stem_width)
-    fallback_stem_width = 0; /* redundant, but avoids a compiler warning */
+    fallback_stem_width = 0; // redundant, but avoids a compiler warning
   if (!have_family_suffix)
     family_suffix = "";
 
 #ifndef BUILD_GUI
 
+  // check SOURE_DATE_EPOCH environment variable
+  const char* source_date_epoch = getenv("SOURCE_DATE_EPOCH");
+  if (source_date_epoch)
+  {
+    char* endptr;
+    errno = 0;
+
+    epoch = strtoull(source_date_epoch, &endptr, 10);
+    if ((errno == ERANGE && (epoch == ULLONG_MAX
+                             || epoch == 0))
+        || (errno != 0
+            && epoch == 0))
+    {
+      fprintf(stderr,
+              "Environment variable `SOURCE_DATE_EPOCH' ignored:\n"
+              "  strtoull: %s\n",
+              strerror(errno));
+      epoch = ULLONG_MAX;
+    }
+    else if (endptr == source_date_epoch)
+    {
+      fprintf(stderr,
+              "Environment variable `SOURCE_DATE_EPOCH' ignored:\n"
+              " No digits were found: %s\n",
+              endptr);
+      epoch = ULLONG_MAX;
+    }
+    else if (*endptr != '\0')
+    {
+      fprintf(stderr,
+              "Environment variable `SOURCE_DATE_EPOCH' ignored:\n"
+              " Trailing garbage: %s\n",
+              endptr);
+      epoch = ULLONG_MAX;
+    }
+    else if (epoch > ULONG_MAX)
+    {
+      fprintf(stderr,
+              "Environment variable `SOURCE_DATE_EPOCH' ignored\n"
+              " value must be smaller than or equal to %lu\n"
+              " but was found to be %llu\n",
+              ULONG_MAX, epoch);
+      epoch = ULLONG_MAX;
+    }
+  }
+
   if (!isatty(fileno(stderr)) && !debug)
     setvbuf(stderr, (char*)NULL, _IONBF, BUFSIZ);
 
@@ -906,10 +1183,10 @@ main(int argc,
 
   if (symbol
       && have_fallback_stem_width
-      && !strcmp(fallback_script, "none"))
+      && fallback_scaling)
     fprintf(stderr,
             "Warning: Setting a fallback stem width for a symbol font\n"
-            "         without setting a fallback script has no effect\n");
+            "         with fallback scaling only has no effect\n");
 
   if (const char* pos = check_family_suffix(family_suffix))
   {
@@ -948,6 +1225,9 @@ main(int argc,
     in = stdin;
   }
 
+  if (show_TTFA_info)
+    display_TTFA(in); // this function doesn't return
+
   FILE* out;
   if (num_args > 1)
   {
@@ -982,7 +1262,8 @@ main(int argc,
     if (!control)
     {
       fprintf(stderr,
-              "The following error occurred while open control file `%s':\n"
+              "The following error occurred"
+                " while opening control file `%s':\n"
               "\n"
               "  %s\n",
               control_name, strerror(errno));
@@ -992,6 +1273,61 @@ main(int argc,
   else
     control = NULL;
 
+  FILE* reference = NULL;
+  if (reference_name)
+  {
+    reference = fopen(reference_name, "r");
+    if (!reference)
+    {
+      fprintf(stderr,
+              "The following error occurred"
+                " while opening reference font `%s':\n"
+              "\n"
+              "  %s\n",
+              reference_name, strerror(errno));
+      exit(EXIT_FAILURE);
+    }
+
+    if (reference_index != 0)
+    {
+      FT_Library library = NULL;
+      FT_Face face = NULL;
+      FT_Error error;
+
+      error = FT_Init_FreeType(&library);
+      if (error)
+      {
+        fprintf(stderr, "Can't initialize FreeType library:\n"
+                        "%s\n",
+                        FT_get_error_message(error));
+        exit(EXIT_FAILURE);
+      }
+
+      error = FT_New_Face(library, reference_name, -1, &face);
+      if (error)
+      {
+        fprintf(stderr, "Can't check number of faces in reference font:\n"
+                        "%s\n",
+                        FT_get_error_message(error));
+        exit(EXIT_FAILURE);
+      }
+
+      if (reference_index < 0
+          || reference_index >= face->num_faces)
+      {
+        fprintf(stderr, "Face index for reference font must be"
+                          " in the range [0;%ld]\n",
+                        face->num_faces - 1);
+        exit(EXIT_FAILURE);
+      }
+
+      FT_Done_Face(face);
+      FT_Done_FreeType(library);
+    }
+  }
+  else
+    reference = NULL;
+
   Progress_Data progress_data = {-1, 1, 0};
   Error_Data error_data = {control_name};
   Info_Data info_data;
@@ -1007,6 +1343,8 @@ main(int argc,
   info_data.info_string_wide_len = 0;
 
   info_data.control_name = control_name;
+  info_data.reference_name = reference_name;
+  info_data.reference_index = reference_index;
 
   info_data.hinting_range_min = hinting_range_min;
   info_data.hinting_range_max = hinting_range_max;
@@ -1025,6 +1363,7 @@ main(int argc,
   info_data.family_data_head = NULL;
   info_data.fallback_stem_width = fallback_stem_width;
   info_data.symbol = symbol;
+  info_data.fallback_scaling = fallback_scaling;
   info_data.TTFA_info = TTFA_info;
 
   strncpy(info_data.default_script,
@@ -1054,6 +1393,7 @@ main(int argc,
 
   TA_Error error =
     TTF_autohint("in-file, out-file, control-file,"
+                 "reference-file, reference-index, reference-name,"
                  "hinting-range-min, hinting-range-max, hinting-limit,"
                  "gray-strong-stem-width, gdi-cleartype-strong-stem-width,"
                  "dw-cleartype-strong-stem-width,"
@@ -1063,9 +1403,11 @@ main(int argc,
                  "ignore-restrictions, windows-compatibility,"
                  "adjust-subglyphs, hint-composites,"
                  "increase-x-height, x-height-snapping-exceptions,"
-                 "fallback-stem-width, default-script, fallback-script,"
-                 "symbol, dehint, debug, TTFA-info",
+                 "fallback-stem-width, default-script,"
+                 "fallback-script, fallback-scaling,"
+                 "symbol, dehint, debug, TTFA-info, epoch",
                  in, out, control,
+                 reference, reference_index, reference_name,
                  hinting_range_min, hinting_range_max, hinting_limit,
                  gray_strong_stem_width, gdi_cleartype_strong_stem_width,
                  dw_cleartype_strong_stem_width,
@@ -1075,8 +1417,9 @@ main(int argc,
                  ignore_restrictions, windows_compatibility,
                  adjust_subglyphs, hint_composites,
                  increase_x_height, x_height_snapping_exceptions_string,
-                 fallback_stem_width, default_script, fallback_script,
-                 symbol, dehint, debug, TTFA_info);
+                 fallback_stem_width, default_script,
+                 fallback_script, fallback_scaling,
+                 symbol, dehint, debug, TTFA_info, epoch);
 
   if (!no_info)
   {
@@ -1090,6 +1433,8 @@ main(int argc,
     fclose(out);
   if (control)
     fclose(control);
+  if (reference)
+    fclose(reference);
 
   exit(error ? EXIT_FAILURE : EXIT_SUCCESS);
 
@@ -1101,7 +1446,7 @@ main(int argc,
   char** new_argv = new char*[new_argc];
 
   // construct new argc and argv variables from collected data
-  for (int i = 0; i < new_argc; i++)
+  for (unsigned int i = 0; i < (unsigned int)new_argc; i++)
     new_argv[i] = const_cast<char*>(new_arg_string[i].data());
 
   QApplication app(new_argc, new_argv);
@@ -1140,8 +1485,8 @@ main(int argc,
                    x_height_snapping_exceptions_string, fallback_stem_width,
                    ignore_restrictions, windows_compatibility, adjust_subglyphs,
                    hint_composites, no_info, detailed_info,
-                   default_script, fallback_script, family_suffix,
-                   symbol, dehint, TTFA_info);
+                   default_script, fallback_script, fallback_scaling,
+                   family_suffix, symbol, dehint, TTFA_info);
 
     dummy.move(-50000, -50000);
     dummy.show();
@@ -1160,8 +1505,8 @@ main(int argc,
                x_height_snapping_exceptions_string, fallback_stem_width,
                ignore_restrictions, windows_compatibility, adjust_subglyphs,
                hint_composites, no_info, detailed_info,
-               default_script, fallback_script, family_suffix,
-               symbol, dehint, TTFA_info);
+               default_script, fallback_script, fallback_scaling,
+               family_suffix, symbol, dehint, TTFA_info);
   gui.show();
 
   return app.exec();
