@@ -1,7 +1,7 @@
 /* tabytecode.c */
 
 /*
- * Copyright (C) 2011-2015 by Werner Lemberg.
+ * Copyright (C) 2011-2017 by Werner Lemberg.
  *
  * This file is part of the ttfautohint library, and may only be used,
  * modified, and distributed under the terms given in `COPYING'.  By
@@ -149,6 +149,7 @@ typedef struct Recorder_
   /* to split them into 24-26 and 0-2 */
   FT_UShort* wrap_around_segments;
   FT_UShort num_wrap_around_segments;
+  FT_Bool wrap_around_segments_initialized;
 
   FT_UShort num_stack_elements; /* the necessary stack depth so far */
 
@@ -158,8 +159,10 @@ typedef struct Recorder_
   ip_on_points ip_on_points_head;
   ip_between_points ip_between_points_head;
 
-  FT_UShort num_strong_points;
-  FT_UShort num_segments;
+  /* we omit one-point segments not part of an edge, */
+  /* thus we have to adjust indices into the `segments' array */
+  FT_UShort* segment_map;
+  FT_Bool segment_map_initialized;
 } Recorder;
 
 
@@ -181,7 +184,7 @@ FT_Byte ttfautohint_glyph_bytecode[7] =
 
 
 /* if we have y delta exceptions before IUP_y, this code gets inserted */
-FT_Byte ins_extra_delta_exceptions[4] =
+static FT_Byte ins_extra_delta_exceptions[4] =
 {
 
   /* tell bci_{scale,scale_composite,hint}_glyph to not call IUP_y */
@@ -194,7 +197,7 @@ FT_Byte ins_extra_delta_exceptions[4] =
 
 
 /* if we have a non-base glyph, this code gets inserted */
-FT_Byte ins_extra_ignore_std_width[4] =
+static FT_Byte ins_extra_ignore_std_width[4] =
 {
 
   /* tell bci_{smooth,strong}_stem_width to ignore std_width */
@@ -311,6 +314,9 @@ TA_optimize_push(FT_Byte* buf,
     pos[2] = NULL;
   }
 
+  /* only needed for the algorithm loops below */
+  pos[3] = NULL;
+
   /* there are at least two NPUSHB instructions */
   /* (one of them directly at the start) */
   sizes[0] = *(pos[0] + 1);
@@ -414,6 +420,25 @@ TA_adjust_point_index(Recorder* recorder,
 }
 
 
+/* We omit one-point segments not part of an edge, */
+/* thus we have to adjust indices into the `segments' array. */
+/* If `segment' is NULL, the number of all segments */
+/* without non-edge one-point segments is returned. */
+
+static FT_UShort
+TA_get_segment_index(TA_Segment segment,
+                     Recorder* recorder)
+{
+  FONT* font = recorder->font;
+  TA_GlyphHints hints = &font->loader->hints;
+  TA_AxisHints axis = &hints->axis[TA_DIMENSION_VERT];
+
+
+  return segment ? recorder->segment_map[segment - axis->segments]
+                 : recorder->segment_map[axis->num_segments];
+}
+
+
 /* we store the segments in the storage area; */
 /* each segment record consists of the first and last point */
 
@@ -455,7 +480,7 @@ TA_sfnt_build_glyph_segments(SFNT* sfnt,
 
 
   seg_limit = segments + axis->num_segments;
-  num_segments = (FT_UShort)axis->num_segments;
+  num_segments = TA_get_segment_index(NULL, recorder);
 
   /* to pack the data in the bytecode more tightly, */
   /* we store up to the first nine segments in nibbles if possible, */
@@ -468,6 +493,9 @@ TA_sfnt_build_glyph_segments(SFNT* sfnt,
     FT_UInt last = (FT_UInt)(seg->last - points);
 
 
+    if (TA_get_segment_index(seg, recorder) == 0xFFFF)
+      continue;
+
     first = TA_adjust_point_index(recorder, first);
     last = TA_adjust_point_index(recorder, last);
 
@@ -513,7 +541,8 @@ TA_sfnt_build_glyph_segments(SFNT* sfnt,
   *(arg--) = num_segments;
 
   base = 0;
-  for (seg = segments; seg < segments + num_packed_segments; seg++)
+  n = 0;
+  for (seg = segments; seg < seg_limit; seg++)
   {
     FT_UInt first = (FT_UInt)(seg->first - points);
     FT_UInt last = (FT_UInt)(seg->last - points);
@@ -521,6 +550,11 @@ TA_sfnt_build_glyph_segments(SFNT* sfnt,
     FT_UInt high_nibble;
 
 
+    if (TA_get_segment_index(seg, recorder) == 0xFFFF)
+      continue;
+    if (n >= num_packed_segments)
+      break;
+
     first = TA_adjust_point_index(recorder, first);
     last = TA_adjust_point_index(recorder, last);
 
@@ -530,14 +564,18 @@ TA_sfnt_build_glyph_segments(SFNT* sfnt,
     *(arg--) = 16 * high_nibble + low_nibble;
 
     base = last;
+    n++;
   }
 
-  for (seg = segments + num_packed_segments; seg < seg_limit; seg++)
+  for (; seg < seg_limit; seg++)
   {
     FT_UInt first = (FT_UInt)(seg->first - points);
     FT_UInt last = (FT_UInt)(seg->last - points);
 
 
+    if (TA_get_segment_index(seg, recorder) == 0xFFFF)
+      continue;
+
     *(arg--) = TA_adjust_point_index(recorder, first);
     *(arg--) = TA_adjust_point_index(recorder, last);
 
@@ -578,6 +616,9 @@ TA_sfnt_build_glyph_segments(SFNT* sfnt,
     FT_UInt last = (FT_UInt)(seg->last - points);
 
 
+    if (TA_get_segment_index(seg, recorder) == 0xFFFF)
+      continue;
+
     if (first > last)
     {
       for (n = 0; n < outline.n_contours; n++)
@@ -1206,8 +1247,8 @@ TA_font_build_subglyph_shifter(FONT* font,
 
   TA_GlyphLoader gloader = font->loader->gloader;
 
-  TA_SubGlyph subglyphs = gloader->base.subglyphs;
-  TA_SubGlyph subglyph_limit = subglyphs + gloader->base.num_subglyphs;
+  TA_SubGlyph subglyphs = gloader->current.subglyphs;
+  TA_SubGlyph subglyph_limit = subglyphs + gloader->current.num_subglyphs;
   TA_SubGlyph subglyph;
 
   FT_Int curr_contour = 0;
@@ -1304,7 +1345,6 @@ TA_build_point_hints(Recorder* recorder,
                      TA_GlyphHints hints)
 {
   TA_AxisHints axis = &hints->axis[TA_DIMENSION_VERT];
-  TA_Segment segments = axis->segments;
   TA_Edge edges = axis->edges;
 
   FT_Byte* p = recorder->hints_record.buf;
@@ -1343,16 +1383,18 @@ TA_build_point_hints(Recorder* recorder,
   if (i)
   {
     TA_Edge edge;
+    FT_UShort edge_first_idx;
 
 
     recorder->hints_record.num_actions++;
 
     edge = edges;
+    edge_first_idx = TA_get_segment_index(edge->first, recorder);
 
     *(p++) = 0;
     *(p++) = (FT_Byte)ta_ip_before + ACTION_OFFSET;
-    *(p++) = HIGH(edge->first - segments);
-    *(p++) = LOW(edge->first - segments);
+    *(p++) = HIGH(edge_first_idx);
+    *(p++) = LOW(edge_first_idx);
     *(p++) = HIGH(i);
     *(p++) = LOW(i);
 
@@ -1389,16 +1431,18 @@ TA_build_point_hints(Recorder* recorder,
   if (i)
   {
     TA_Edge edge;
+    FT_UShort edge_first_idx;
 
 
     recorder->hints_record.num_actions++;
 
     edge = edges + axis->num_edges - 1;
+    edge_first_idx = TA_get_segment_index(edge->first, recorder);
 
     *(p++) = 0;
     *(p++) = (FT_Byte)ta_ip_after + ACTION_OFFSET;
-    *(p++) = HIGH(edge->first - segments);
-    *(p++) = LOW(edge->first - segments);
+    *(p++) = HIGH(edge_first_idx);
+    *(p++) = LOW(edge_first_idx);
     *(p++) = HIGH(i);
     *(p++) = LOW(i);
 
@@ -1455,12 +1499,14 @@ TA_build_point_hints(Recorder* recorder,
     {
       Node2* edge_node;
       TA_Edge edge;
+      FT_UShort edge_first_idx;
 
 
       edge = edges + on_node->edge;
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
 
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
 
       /* save current position */
       edge_node = on_node;
@@ -1548,15 +1594,19 @@ TA_build_point_hints(Recorder* recorder,
       Node3* edge_pair_node;
       TA_Edge before;
       TA_Edge after;
+      FT_UShort before_first_idx;
+      FT_UShort after_first_idx;
 
 
       before = edges + between_node->before_edge;
       after = edges + between_node->after_edge;
+      before_first_idx = TA_get_segment_index(before->first, recorder);
+      after_first_idx = TA_get_segment_index(after->first, recorder);
 
-      *(p++) = HIGH(after->first - segments);
-      *(p++) = LOW(after->first - segments);
-      *(p++) = HIGH(before->first - segments);
-      *(p++) = LOW(before->first - segments);
+      *(p++) = HIGH(after_first_idx);
+      *(p++) = LOW(after_first_idx);
+      *(p++) = HIGH(before_first_idx);
+      *(p++) = LOW(before_first_idx);
 
       /* save current position */
       edge_pair_node = between_node;
@@ -1821,18 +1871,19 @@ TA_free_hints_records(Hints_Record* hints_records,
 
 static FT_Byte*
 TA_hints_recorder_handle_segments(FT_Byte* bufp,
-                                  TA_AxisHints axis,
+                                  Recorder* recorder,
                                   TA_Edge edge,
                                   FT_UShort* wraps)
 {
-  TA_Segment segments = axis->segments;
   TA_Segment seg;
   FT_UShort seg_idx;
   FT_UShort num_segs = 0;
   FT_UShort* wrap;
+  FT_UShort num_segments;
 
 
-  seg_idx = (FT_UShort)(edge->first - segments);
+  seg_idx = TA_get_segment_index(edge->first, recorder);
+  num_segments = TA_get_segment_index(NULL, recorder);
 
   /* we store everything as 16bit numbers */
   *(bufp++) = HIGH(seg_idx);
@@ -1868,14 +1919,14 @@ TA_hints_recorder_handle_segments(FT_Byte* bufp,
       wrap++;
     }
 
-    *(bufp++) = HIGH(axis->num_segments + (wrap - wraps));
-    *(bufp++) = LOW(axis->num_segments + (wrap - wraps));
+    *(bufp++) = HIGH(num_segments + (wrap - wraps));
+    *(bufp++) = LOW(num_segments + (wrap - wraps));
   }
 
   seg = edge->first->edge_next;
   while (seg != edge->first)
   {
-    seg_idx = (FT_UShort)(seg - segments);
+    seg_idx = TA_get_segment_index(seg, recorder);
 
     *(bufp++) = HIGH(seg_idx);
     *(bufp++) = LOW(seg_idx);
@@ -1890,8 +1941,8 @@ TA_hints_recorder_handle_segments(FT_Byte* bufp,
         wrap++;
       }
 
-      *(bufp++) = HIGH(axis->num_segments + (wrap - wraps));
-      *(bufp++) = LOW(axis->num_segments + (wrap - wraps));
+      *(bufp++) = HIGH(num_segments + (wrap - wraps));
+      *(bufp++) = LOW(num_segments + (wrap - wraps));
     }
 
     seg = seg->edge_next;
@@ -1913,21 +1964,62 @@ TA_hints_recorder(TA_Action action,
 {
   TA_AxisHints axis = &hints->axis[dim];
   TA_Edge edges = axis->edges;
-  TA_Segment segments = axis->segments;
   TA_Point points = hints->points;
 
+  TA_Segment segments = axis->segments;
+  TA_Segment seg_limit = segments + axis->num_segments;
+  TA_Segment seg;
+
   Recorder* recorder = (Recorder*)hints->user;
   SFNT* sfnt = recorder->sfnt;
   FONT* font = recorder->font;
   FT_UShort* wraps = recorder->wrap_around_segments;
   FT_Byte* p = recorder->hints_record.buf;
 
-  FT_UInt style = font->loader->metrics->style_class->style;
+  TA_StyleClass style_class = font->loader->metrics->style_class;
+  TA_ScriptClass script_class = ta_script_classes[style_class->script];
+
+  FT_UInt style = style_class->style;
+  FT_Bool top_to_bottom_hinting = script_class->top_to_bottom_hinting;
 
 
   if (dim == TA_DIMENSION_HORZ)
     return;
 
+  if (!recorder->segment_map_initialized)
+  {
+    FT_UShort* segment_map = recorder->segment_map;
+    FT_UShort idx;
+
+
+    idx = 0;
+    for (seg = segments; seg < seg_limit; seg++)
+    {
+      if (seg->edge)
+        *segment_map = idx++;
+      else
+        *segment_map = 0xFFFF;
+
+      segment_map++;
+    }
+    *segment_map = idx;
+
+    recorder->segment_map_initialized = 1;
+  }
+
+  if (!recorder->wrap_around_segments_initialized)
+  {
+    FT_UShort* wrap_around_segment;
+
+
+    wrap_around_segment = recorder->wrap_around_segments;
+    for (seg = segments; seg < seg_limit; seg++)
+      if (seg->first > seg->last)
+        *(wrap_around_segment++) = TA_get_segment_index(seg, recorder);
+
+    recorder->wrap_around_segments_initialized = 1;
+  }
+
   /* we collect point hints for later processing */
   switch (action)
   {
@@ -2014,8 +2106,8 @@ TA_hints_recorder(TA_Action action,
     break;
   }
 
-  /* some enum values correspond to four or eight bytecode functions; */
-  /* if the value is n, the function numbers are n, ..., n+7, */
+  /* some enum values correspond to 4, 7, 8, or 12 bytecode functions; */
+  /* if the value is n, the function numbers are n, ..., n+11, */
   /* to be differentiated with flags */
 
   switch (action)
@@ -2024,19 +2116,24 @@ TA_hints_recorder(TA_Action action,
     {
       TA_Edge base_edge = (TA_Edge)arg1;
       TA_Edge stem_edge = arg2;
+      FT_UShort base_first_idx;
+      FT_UShort stem_first_idx;
 
 
+      base_first_idx = TA_get_segment_index(base_edge->first, recorder);
+      stem_first_idx = TA_get_segment_index(stem_edge->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET
                + ((stem_edge->flags & TA_EDGE_SERIF) != 0)
                + 2 * ((base_edge->flags & TA_EDGE_ROUND) != 0);
 
-      *(p++) = HIGH(base_edge->first - segments);
-      *(p++) = LOW(base_edge->first - segments);
-      *(p++) = HIGH(stem_edge->first - segments);
-      *(p++) = LOW(stem_edge->first - segments);
+      *(p++) = HIGH(base_first_idx);
+      *(p++) = LOW(base_first_idx);
+      *(p++) = HIGH(stem_first_idx);
+      *(p++) = LOW(stem_first_idx);
 
-      p = TA_hints_recorder_handle_segments(p, axis, stem_edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, stem_edge, wraps);
     }
     break;
 
@@ -2044,19 +2141,24 @@ TA_hints_recorder(TA_Action action,
     {
       TA_Edge edge = (TA_Edge)arg1;
       TA_Edge edge2 = arg2;
+      FT_UShort edge_first_idx;
+      FT_UShort edge2_first_idx;
 
 
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
+      edge2_first_idx = TA_get_segment_index(edge2->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET
                + ((edge2->flags & TA_EDGE_SERIF) != 0)
                + 2 * ((edge->flags & TA_EDGE_ROUND) != 0);
 
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
-      *(p++) = HIGH(edge2->first - segments);
-      *(p++) = LOW(edge2->first - segments);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
+      *(p++) = HIGH(edge2_first_idx);
+      *(p++) = LOW(edge2_first_idx);
 
-      p = TA_hints_recorder_handle_segments(p, axis, edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge, wraps);
     }
     break;
 
@@ -2065,26 +2167,39 @@ TA_hints_recorder(TA_Action action,
       TA_Edge edge = (TA_Edge)arg1;
       TA_Edge edge2 = arg2;
       TA_Edge edge_minus_one = lower_bound;
+      FT_UShort edge_first_idx;
+      FT_UShort edge2_first_idx;
 
 
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
+      edge2_first_idx = TA_get_segment_index(edge2->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET
                + ((edge2->flags & TA_EDGE_SERIF) != 0)
                + 2 * ((edge->flags & TA_EDGE_ROUND) != 0)
-               + 4 * (edge_minus_one != NULL);
+               + 4 * (edge_minus_one != NULL) /* `bound' */
+               + 4 * (edge_minus_one != NULL
+                      && top_to_bottom_hinting); /* `down' */
 
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
-      *(p++) = HIGH(edge2->first - segments);
-      *(p++) = LOW(edge2->first - segments);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
+      *(p++) = HIGH(edge2_first_idx);
+      *(p++) = LOW(edge2_first_idx);
 
       if (edge_minus_one)
       {
-        *(p++) = HIGH(edge_minus_one->first - segments);
-        *(p++) = LOW(edge_minus_one->first - segments);
+        FT_UShort edge_minus_one_first_idx;
+
+
+        edge_minus_one_first_idx = TA_get_segment_index(
+                                     edge_minus_one->first, recorder);
+
+        *(p++) = HIGH(edge_minus_one_first_idx);
+        *(p++) = LOW(edge_minus_one_first_idx);
       }
 
-      p = TA_hints_recorder_handle_segments(p, axis, edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge, wraps);
     }
     break;
 
@@ -2092,13 +2207,18 @@ TA_hints_recorder(TA_Action action,
     {
       TA_Edge edge = (TA_Edge)arg1;
       TA_Edge blue = arg2;
+      FT_UShort blue_first_idx;
+      FT_UShort edge_first_idx;
 
 
+      blue_first_idx = TA_get_segment_index(blue->first, recorder);
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET;
 
-      *(p++) = HIGH(blue->first - segments);
-      *(p++) = LOW(blue->first - segments);
+      *(p++) = HIGH(blue_first_idx);
+      *(p++) = LOW(blue_first_idx);
 
       if (edge->best_blue_is_shoot)
       {
@@ -2111,10 +2231,10 @@ TA_hints_recorder(TA_Action action,
         *(p++) = LOW(CVT_BLUE_REFS_OFFSET(style) + edge->best_blue_idx);
       }
 
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
 
-      p = TA_hints_recorder_handle_segments(p, axis, edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge, wraps);
     }
     break;
 
@@ -2123,35 +2243,51 @@ TA_hints_recorder(TA_Action action,
       TA_Edge edge = (TA_Edge)arg1;
       TA_Edge edge2 = arg2;
       TA_Edge edge_minus_one = lower_bound;
+      FT_UShort edge_first_idx;
+      FT_UShort edge2_first_idx;
 
 
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
+      edge2_first_idx = TA_get_segment_index(edge2->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET
                + ((edge2->flags & TA_EDGE_SERIF) != 0)
                + 2 * ((edge->flags & TA_EDGE_ROUND) != 0)
-               + 4 * (edge_minus_one != NULL);
+               + 4 * (edge_minus_one != NULL) /* `bound' */
+               + 4 * (edge_minus_one != NULL
+                      && top_to_bottom_hinting); /* `down' */
 
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
-      *(p++) = HIGH(edge2->first - segments);
-      *(p++) = LOW(edge2->first - segments);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
+      *(p++) = HIGH(edge2_first_idx);
+      *(p++) = LOW(edge2_first_idx);
 
       if (edge_minus_one)
       {
-        *(p++) = HIGH(edge_minus_one->first - segments);
-        *(p++) = LOW(edge_minus_one->first - segments);
+        FT_UShort edge_minus_one_first_idx;
+
+
+        edge_minus_one_first_idx = TA_get_segment_index(
+                                     edge_minus_one->first, recorder);
+
+        *(p++) = HIGH(edge_minus_one_first_idx);
+        *(p++) = LOW(edge_minus_one_first_idx);
       }
 
-      p = TA_hints_recorder_handle_segments(p, axis, edge, wraps);
-      p = TA_hints_recorder_handle_segments(p, axis, edge2, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge2, wraps);
     }
     break;
 
   case ta_blue:
     {
       TA_Edge edge = (TA_Edge)arg1;
+      FT_UShort edge_first_idx;
 
 
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET;
 
@@ -2166,10 +2302,10 @@ TA_hints_recorder(TA_Action action,
         *(p++) = LOW(CVT_BLUE_REFS_OFFSET(style) + edge->best_blue_idx);
       }
 
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
 
-      p = TA_hints_recorder_handle_segments(p, axis, edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge, wraps);
     }
     break;
 
@@ -2177,30 +2313,49 @@ TA_hints_recorder(TA_Action action,
     {
       TA_Edge serif = (TA_Edge)arg1;
       TA_Edge base = serif->serif;
+      FT_UShort serif_first_idx;
+      FT_UShort base_first_idx;
 
 
+      serif_first_idx = TA_get_segment_index(serif->first, recorder);
+      base_first_idx = TA_get_segment_index(base->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET
                + (lower_bound != NULL)
-               + 2 * (upper_bound != NULL);
+               + 2 * (upper_bound != NULL)
+               + 3 * ((lower_bound != NULL || upper_bound != NULL)
+                      && top_to_bottom_hinting); /* `down' */
 
-      *(p++) = HIGH(serif->first - segments);
-      *(p++) = LOW(serif->first - segments);
-      *(p++) = HIGH(base->first - segments);
-      *(p++) = LOW(base->first - segments);
+      *(p++) = HIGH(serif_first_idx);
+      *(p++) = LOW(serif_first_idx);
+      *(p++) = HIGH(base_first_idx);
+      *(p++) = LOW(base_first_idx);
 
       if (lower_bound)
       {
-        *(p++) = HIGH(lower_bound->first - segments);
-        *(p++) = LOW(lower_bound->first - segments);
+        FT_UShort lower_bound_first_idx;
+
+
+        lower_bound_first_idx = TA_get_segment_index(lower_bound->first,
+                                                     recorder);
+
+        *(p++) = HIGH(lower_bound_first_idx);
+        *(p++) = LOW(lower_bound_first_idx);
       }
       if (upper_bound)
       {
-        *(p++) = HIGH(upper_bound->first - segments);
-        *(p++) = LOW(upper_bound->first - segments);
+        FT_UShort upper_bound_first_idx;
+
+
+        upper_bound_first_idx = TA_get_segment_index(upper_bound->first,
+                                                     recorder);
+
+        *(p++) = HIGH(upper_bound_first_idx);
+        *(p++) = LOW(upper_bound_first_idx);
       }
 
-      p = TA_hints_recorder_handle_segments(p, axis, serif, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, serif, wraps);
     }
     break;
 
@@ -2208,28 +2363,45 @@ TA_hints_recorder(TA_Action action,
   case ta_serif_link2:
     {
       TA_Edge edge = (TA_Edge)arg1;
+      FT_UShort edge_first_idx;
 
 
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET
                + (lower_bound != NULL)
-               + 2 * (upper_bound != NULL);
+               + 2 * (upper_bound != NULL)
+               + 3 * ((lower_bound != NULL || upper_bound != NULL)
+                      && top_to_bottom_hinting); /* `down' */
 
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
 
       if (lower_bound)
       {
-        *(p++) = HIGH(lower_bound->first - segments);
-        *(p++) = LOW(lower_bound->first - segments);
+        FT_UShort lower_bound_first_idx;
+
+
+        lower_bound_first_idx = TA_get_segment_index(lower_bound->first,
+                                                     recorder);
+
+        *(p++) = HIGH(lower_bound_first_idx);
+        *(p++) = LOW(lower_bound_first_idx);
       }
       if (upper_bound)
       {
-        *(p++) = HIGH(upper_bound->first - segments);
-        *(p++) = LOW(upper_bound->first - segments);
+        FT_UShort upper_bound_first_idx;
+
+
+        upper_bound_first_idx = TA_get_segment_index(upper_bound->first,
+                                                     recorder);
+
+        *(p++) = HIGH(upper_bound_first_idx);
+        *(p++) = LOW(upper_bound_first_idx);
       }
 
-      p = TA_hints_recorder_handle_segments(p, axis, edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge, wraps);
     }
     break;
 
@@ -2238,32 +2410,53 @@ TA_hints_recorder(TA_Action action,
       TA_Edge edge = (TA_Edge)arg1;
       TA_Edge before = arg2;
       TA_Edge after = arg3;
+      FT_UShort before_first_idx;
+      FT_UShort edge_first_idx;
+      FT_UShort after_first_idx;
 
 
+      before_first_idx = TA_get_segment_index(before->first, recorder);
+      edge_first_idx = TA_get_segment_index(edge->first, recorder);
+      after_first_idx = TA_get_segment_index(after->first, recorder);
+
       *(p++) = 0;
       *(p++) = (FT_Byte)action + ACTION_OFFSET
                + (lower_bound != NULL)
-               + 2 * (upper_bound != NULL);
+               + 2 * (upper_bound != NULL)
+               + 3 * ((lower_bound != NULL || upper_bound != NULL)
+                      && top_to_bottom_hinting); /* `down' */
 
-      *(p++) = HIGH(before->first - segments);
-      *(p++) = LOW(before->first - segments);
-      *(p++) = HIGH(edge->first - segments);
-      *(p++) = LOW(edge->first - segments);
-      *(p++) = HIGH(after->first - segments);
-      *(p++) = LOW(after->first - segments);
+      *(p++) = HIGH(before_first_idx);
+      *(p++) = LOW(before_first_idx);
+      *(p++) = HIGH(edge_first_idx);
+      *(p++) = LOW(edge_first_idx);
+      *(p++) = HIGH(after_first_idx);
+      *(p++) = LOW(after_first_idx);
 
       if (lower_bound)
       {
-        *(p++) = HIGH(lower_bound->first - segments);
-        *(p++) = LOW(lower_bound->first - segments);
+        FT_UShort lower_bound_first_idx;
+
+
+        lower_bound_first_idx = TA_get_segment_index(lower_bound->first,
+                                                     recorder);
+
+        *(p++) = HIGH(lower_bound_first_idx);
+        *(p++) = LOW(lower_bound_first_idx);
       }
       if (upper_bound)
       {
-        *(p++) = HIGH(upper_bound->first - segments);
-        *(p++) = LOW(upper_bound->first - segments);
+        FT_UShort upper_bound_first_idx;
+
+
+        upper_bound_first_idx = TA_get_segment_index(upper_bound->first,
+                                                     recorder);
+
+        *(p++) = HIGH(upper_bound_first_idx);
+        *(p++) = LOW(upper_bound_first_idx);
       }
 
-      p = TA_hints_recorder_handle_segments(p, axis, edge, wraps);
+      p = TA_hints_recorder_handle_segments(p, recorder, edge, wraps);
     }
     break;
 
@@ -2286,21 +2479,15 @@ TA_init_recorder(Recorder* recorder,
                  TA_GlyphHints hints)
 {
   TA_AxisHints axis = &hints->axis[TA_DIMENSION_VERT];
-  TA_Point points = hints->points;
-  TA_Point point_limit = points + hints->num_points;
-  TA_Point point;
 
   TA_Segment segments = axis->segments;
   TA_Segment seg_limit = segments + axis->num_segments;
   TA_Segment seg;
 
-  FT_UShort num_strong_points = 0;
-  FT_UShort* wrap_around_segment;
 
   recorder->sfnt = sfnt;
   recorder->font = font;
   recorder->glyph = glyph;
-  recorder->num_segments = (FT_UShort)axis->num_segments;
 
   LLRB_INIT(&recorder->ip_before_points_head);
   LLRB_INIT(&recorder->ip_after_points_head);
@@ -2312,6 +2499,17 @@ TA_init_recorder(Recorder* recorder,
   /* no need to clean up allocated arrays in case of error; */
   /* this is handled later by `TA_free_recorder' */
 
+  /* we use segment_map[axis->num_segments] */
+  /* as the total number of mapped segments, so allocate one more element */
+  recorder->segment_map =
+    (FT_UShort*)malloc((size_t)(axis->num_segments + 1) * sizeof (FT_UShort));
+  if (!recorder->segment_map)
+    return FT_Err_Out_Of_Memory;
+
+  /* `segment_map' gets initialized later on, */
+  /* after the first call of `ta_loader_load_glyph' */
+  recorder->segment_map_initialized = 0;
+
   recorder->num_wrap_around_segments = 0;
   for (seg = segments; seg < seg_limit; seg++)
     if (seg->first > seg->last)
@@ -2323,24 +2521,10 @@ TA_init_recorder(Recorder* recorder,
   if (!recorder->wrap_around_segments)
     return FT_Err_Out_Of_Memory;
 
-  wrap_around_segment = recorder->wrap_around_segments;
-  for (seg = segments; seg < seg_limit; seg++)
-    if (seg->first > seg->last)
-      *(wrap_around_segment++) = (FT_UShort)(seg - segments);
-
-  /* get number of strong points */
-  for (point = points; point < point_limit; point++)
-  {
-    /* actually, we need to test `TA_FLAG_TOUCH_Y' also; */
-    /* however, this value isn't known yet */
-    /* (or rather, it can vary between different pixel sizes) */
-    if (point->flags & TA_FLAG_WEAK_INTERPOLATION)
-      continue;
-
-    num_strong_points++;
-  }
-
-  recorder->num_strong_points = num_strong_points;
+  /* `wrap_around_segments' gets initialized later on; */
+  /* it needs function `TA_get_segment_index' which uses data */
+  /* that hasn't been initialized yet either */
+  recorder->wrap_around_segments_initialized = 0;
 
   return FT_Err_Ok;
 }
@@ -2438,6 +2622,7 @@ TA_rewind_recorder(Recorder* recorder,
 static void
 TA_free_recorder(Recorder* recorder)
 {
+  free(recorder->segment_map);
   free(recorder->wrap_around_segments);
 
   TA_rewind_recorder(recorder, NULL, 0);
@@ -2480,7 +2665,9 @@ TA_sfnt_build_glyph_instructions(SFNT* sfnt,
   FT_Int32 load_flags;
   FT_UInt size;
 
-  FT_Byte* pos[3];
+  /* we store only three positions, but it simplifies the algorithm in */
+  /* `TA_optimize_push' if we have one additional element */
+  FT_Byte* pos[4];
 
 #ifdef TA_DEBUG
   int _ta_debug_save;
@@ -2533,15 +2720,12 @@ TA_sfnt_build_glyph_instructions(SFNT* sfnt,
     return error;
 
   /* do nothing if we have an empty glyph */
-  if (!face->glyph->outline.n_contours)
+  if (!(font->loader->gloader->current.num_subglyphs
+        || face->glyph->outline.n_contours))
     return FT_Err_Ok;
 
   hints = &font->loader->hints;
 
-  /* do nothing if the setup delivered the `none_dflt' style only */
-  if (!hints->num_points)
-    return FT_Err_Ok;
-
   /*
    * We allocate a buffer which is certainly large enough
    * to hold all of the created bytecode instructions;
@@ -2567,7 +2751,7 @@ TA_sfnt_build_glyph_instructions(SFNT* sfnt,
     return FT_Err_Out_Of_Memory;
 
   /* handle composite glyph */
-  if (font->loader->gloader->base.num_subglyphs)
+  if (font->loader->gloader->current.num_subglyphs)
   {
     bufp = TA_font_build_subglyph_shifter(font, ins_buf);
     if (!bufp)
@@ -2581,20 +2765,30 @@ TA_sfnt_build_glyph_instructions(SFNT* sfnt,
     goto Done1;
   }
 
-  /* only scale the glyph if the `none_dflt' style has been used */
-  if (font->loader->metrics->style_class == &ta_none_dflt_style_class)
-  {
-    /* since `TA_init_recorder' hasn't been called yet, */
-    /* we manually initialize the `sfnt', `font', and `glyph' fields */
-    recorder.sfnt = sfnt;
-    recorder.font = font;
-    recorder.glyph = glyph;
 
-    bufp = TA_sfnt_build_glyph_scaler(sfnt, &recorder, ins_buf);
-    if (!bufp)
+  if (font->fallback_scaling)
+  {
+    if (font->loader->metrics->style_class == &ta_none_dflt_style_class)
     {
-      error = FT_Err_Out_Of_Memory;
-      goto Err;
+      /* the scaling value of `none_dflt' */
+      /* (this is, hinting without script-specific blue zones) */
+      /* is always 1, which corresponds to a no-op */
+      bufp = ins_buf;
+    }
+    else
+    {
+      /* since `TA_init_recorder' hasn't been called yet, */
+      /* we manually initialize the `sfnt', `font', and `glyph' fields */
+      recorder.sfnt = sfnt;
+      recorder.font = font;
+      recorder.glyph = glyph;
+
+      bufp = TA_sfnt_build_glyph_scaler(sfnt, &recorder, ins_buf);
+      if (!bufp)
+      {
+        error = FT_Err_Out_Of_Memory;
+        goto Err;
+      }
     }
 
     use_gstyle_data = 0;
@@ -2634,6 +2828,13 @@ TA_sfnt_build_glyph_instructions(SFNT* sfnt,
                                     TA_hints_recorder,
                                     (void*)&recorder);
 
+  /*
+   * It is important that we start the loop with the smallest PPEM value
+   * used for hinting, since the number of segments that form an edge can
+   * become smaller for larger PPEM values.  For efficiency, we skip
+   * non-edge one-point segments, and `TA_get_segment_index' would return
+   * wrong indices otherwise.
+   */
   for (size = font->hinting_range_min;
        size <= font->hinting_range_max;
        size++)
