@@ -1,7 +1,7 @@
 /* tabytecode.h */
 
 /*
- * Copyright (C) 2011-2016 by Werner Lemberg.
+ * Copyright (C) 2011-2017 by Werner Lemberg.
  *
  * This file is part of the ttfautohint library, and may only be used,
  * modified, and distributed under the terms given in `COPYING'.  By
@@ -311,7 +311,8 @@
 #define sal_func sal_ref + 1
 #define sal_anchor sal_func + 1
 #define sal_stem_width_function sal_anchor + 1
-#define sal_vwidth_data_offset sal_stem_width_function + 1
+#define sal_base_delta sal_stem_width_function + 1
+#define sal_vwidth_data_offset sal_base_delta + 1
 #define sal_scale sal_vwidth_data_offset + 1
 #define sal_point_min sal_scale + 1
 #define sal_point_max sal_point_min + 1
@@ -324,8 +325,8 @@
 /* bytecode function numbers */
 
 /* 0 */
-#define bci_align_top 0
-#define bci_round bci_align_top + 1
+#define bci_align_x_height 0
+#define bci_round bci_align_x_height + 1
 #define bci_smooth_stem_width bci_round + 1
 #define bci_get_best_width bci_smooth_stem_width + 1
 #define bci_strong_stem_width bci_get_best_width + 1
