@@ -1,7 +1,7 @@
 /* taprep.c */
 
 /*
- * Copyright (C) 2011-2015 by Werner Lemberg.
+ * Copyright (C) 2011-2017 by Werner Lemberg.
  *
  * This file is part of the ttfautohint library, and may only be used,
  * modified, and distributed under the terms given in `COPYING'.  By
@@ -82,7 +82,7 @@ static const unsigned char PREP(test_exception_a) [] =
 
 /* provide scaling factors for all styles */
 
-static const unsigned char PREP(align_top_a) [] =
+static const unsigned char PREP(align_x_height_a) [] =
 {
 
     PUSHB_2,
@@ -98,10 +98,10 @@ static const unsigned char PREP(align_top_a) [] =
 /*    %c, style 0's x height blue zone idx */
 /*    %c, num_used_styles */
 
-static const unsigned char PREP(align_top_b) [] =
+static const unsigned char PREP(align_x_height_b) [] =
 {
 
-      bci_align_top,
+      bci_align_x_height,
     LOOPCALL,
 
 };
@@ -225,27 +225,52 @@ static const unsigned char PREP(store_vwidth_data_e) [] =
 
 static const unsigned char PREP(set_smooth_or_strong_a) [] =
 {
-
   /*
-   * There are two ClearType flavours available on Windows: The older GDI
-   * ClearType, introduced in 2000, and the recent DW ClearType, introduced
-   * in 2008.  The main difference is that the older incarnation behaves
-   * like a B/W renderer along the y axis, while the newer version does
-   * vertical smoothing also.
+   * ttfautohint provides two different functions for stem width computation
+   * and blue zone rounding: `smooth' and `strong'.  The former tries to
+   * align stem widths and blue zones to some discrete, possibly non-integer
+   * values.  The latter snaps everything to integer pixels as much as
+   * possible.
    *
-   * The only possibility to differentiate between GDI and DW ClearType is
-   * testing bit 10 in the GETINFO instruction (with return value in bit 17;
-   * this works for TrueType version >= 38), checking whether sub-pixel
-   * positioning is available.
+   * We test ClearType capabilities to find out which of the two functions
+   * should be used.  Due to the various TrueType interpreter versions this
+   * is quite convoluted.
    *
-   * If GDI ClearType is active, we use different functions for stem width
-   * computation and blue zone rounding that snap to integer pixels as much
-   * as possible.
+   *   interpreter version  action
+   *   ---------------------------------------------------------------------
+   *          <= 35         this version predates ClearType -> smooth
+   *
+   *          36-38         use bit 6 in the GETINFO instruction to check
+   *                        whether ClearType is enabled; if set, we have
+   *                        (old) GDI ClearType -> strong, otherwise
+   *                        grayscale rendering -> smooth
+   *
+   *          39            if ClearType is enabled, use bit 10 in the
+   *                        GETINFO instruction to check whether ClearType
+   *                        sub-pixel positioning is available; if set, we
+   *                        have DW ClearType -> smooth, else GDI ClearType
+   *                        -> strong
+   *
+   *          >= 40         if ClearType is enabled, use bit 11 in the
+   *                        GETINFO instruction to check whether ClearType
+   *                        symmetric rendering is available; if not set,
+   *                        the engine behaves like a B/W renderer along the
+   *                        y axis -> strong, else it does vertical
+   *                        smoothing -> smooth
+   *
+   * ClearType on Windows was introduced in 2000 for the GDI framework (no
+   * symmetric rendering, no sub-pixel positioning).  In 2008, Windows got
+   * the DirectWrite (DW) framework which uses symmetric rendering and
+   * sub-pixel positioning.
+   *
+   * Note that in 2017 GDI on Windows 10 has changed rendering parameters:
+   * it now uses symmetric rendering but no sub-pixel positioning.
+   * Consequently, we treat this as `DW ClearType' also.
    */
 
   /* set default value */
   PUSHB_2,
-    cvtl_use_strong_functions,
+    cvtl_use_strong_functions, /* target: grayscale rendering */
 
 };
 
@@ -262,8 +287,7 @@ static const unsigned char PREP(set_smooth_or_strong_b) [] =
     0x01,
   GETINFO,
 
-  /* `GDI ClearType': */
-  /* version >= 36 and version < 38, ClearType enabled */
+  /* `(old) GDI ClearType': version >= 36 || version <= 38 */
   LTEQ,
   IF,
     /* check whether ClearType is enabled (bit 6) */
@@ -272,7 +296,7 @@ static const unsigned char PREP(set_smooth_or_strong_b) [] =
     GETINFO,
     IF,
       PUSHB_2,
-        cvtl_use_strong_functions,
+        cvtl_use_strong_functions, /* target: GDI ClearType */
 };
 
 /*      %c, either 0 or 100 */
@@ -284,31 +308,21 @@ static const unsigned char PREP(set_smooth_or_strong_c) [] =
 
       /* get rasterizer version (bit 0) */
       PUSHB_2,
-        38,
+        40,
         0x01,
       GETINFO,
 
-      /* `DW ClearType': */
-      /* version >= 38, sub-pixel positioning is enabled */
+      /* `DW ClearType': version >= 40 */
       LTEQ,
       IF,
-        /* check whether sub-pixel positioning is enabled (bit 10) -- */
-        /* due to a bug in FreeType 2.5.0 and earlier, */
-        /* bit 6 must be set also to get the correct information, */
-        /* so we test that both return values (in bits 13 and 17) are set */
-        PUSHW_3,
-          0x08, /* bits 13 and 17 shifted by 6 bits */
-          0x80,
-          0x00, /* we do `MUL' with value 1, */
-          0x01, /* which is essentially a division by 64 */
-          0x04, /* bits 6 and 10 */
-          0x40,
+        /* check whether symmetric rendering is enabled (bit 11) */
+        PUSHW_1,
+          0x08,
+          0x00,
         GETINFO,
-        MUL,
-        EQ,
         IF,
           PUSHB_2,
-            cvtl_use_strong_functions,
+            cvtl_use_strong_functions, /* target: DirectWrite ClearType */
 
 };
 
@@ -319,6 +333,45 @@ static const unsigned char PREP(set_smooth_or_strong_d) [] =
 
           WCVTP,
         EIF,
+
+      ELSE,
+        /* get rasterizer version (bit 0) */
+        PUSHB_2,
+          39,
+          0x01,
+        GETINFO,
+
+        /* `DW ClearType': version == 39 */
+        LTEQ,
+        IF,
+          /* check whether sub-pixel positioning is enabled (bit 10) -- */
+          /* due to a bug in FreeType 2.5.0 and earlier, */
+          /* bit 6 must be set also to get the correct information, */
+          /* so we test that both return values (in bits 13 and 17) are set */
+          PUSHW_3,
+            0x08, /* bits 13 and 17 right-shifted by 6 bits */
+            0x80,
+            0x00, /* we do `MUL' with value 1, */
+            0x01, /* which is essentially a division by 64 */
+            0x04, /* bits 6 and 10 */
+            0x40,
+          GETINFO,
+          MUL,
+          EQ,
+          IF,
+            PUSHB_2,
+              cvtl_use_strong_functions, /* target: DirectWrite ClearType */
+
+};
+
+/*            %c, either 0 or 100 */
+
+static const unsigned char PREP(set_smooth_or_strong_e) [] =
+{
+
+            WCVTP,
+          EIF,
+        EIF,
       EIF,
     EIF,
   EIF,
@@ -621,10 +674,10 @@ TA_table_build_prep(FT_Byte** prep,
   if (font->x_height_snapping_exceptions)
     buf_new_len += sizeof (PREP(test_exception_a));
 
-  buf_new_len += sizeof (PREP(align_top_a))
+  buf_new_len += sizeof (PREP(align_x_height_a))
                  + (num_used_styles > 6 ? num_used_styles + 3
                                         : num_used_styles + 2)
-                 + sizeof (PREP(align_top_b));
+                 + sizeof (PREP(align_x_height_b));
   buf_new_len += sizeof (PREP(loop_cvt_a))
                  + (num_used_styles > 3 ? 2 * num_used_styles + 3
                                         : 2 * num_used_styles + 2)
@@ -653,7 +706,9 @@ TA_table_build_prep(FT_Byte** prep,
                  + 1
                  + sizeof (PREP(set_smooth_or_strong_c))
                  + 1
-                 + sizeof (PREP(set_smooth_or_strong_d));
+                 + sizeof (PREP(set_smooth_or_strong_d))
+                 + 1
+                 + sizeof (PREP(set_smooth_or_strong_e));
   buf_new_len += (num_used_styles > 3 ? 2 * num_used_styles + 3
                                       : 2 * num_used_styles + 2)
                  + sizeof (PREP(round_blues));
@@ -695,7 +750,7 @@ TA_table_build_prep(FT_Byte** prep,
   if (font->x_height_snapping_exceptions)
     COPY_PREP(test_exception_a);
 
-  COPY_PREP(align_top_a);
+  COPY_PREP(align_x_height_a);
   if (num_used_styles > 6)
   {
     BCI(NPUSHB);
@@ -714,7 +769,7 @@ TA_table_build_prep(FT_Byte** prep,
                   : (unsigned char)CVT_X_HEIGHT_BLUE_OFFSET(i);
   }
   *(bufp++) = num_used_styles;
-  COPY_PREP(align_top_b);
+  COPY_PREP(align_x_height_b);
 
   COPY_PREP(loop_cvt_a);
   if (num_used_styles > 3)
@@ -735,7 +790,10 @@ TA_table_build_prep(FT_Byte** prep,
     *(bufp++) = (unsigned char)(
                   1
                   + CVT_VERT_WIDTHS_SIZE(i)
-                  + (CVT_BLUES_SIZE(i) > 1 ? CVT_BLUES_SIZE(i) - 2 : 0));
+                  + (CVT_BLUES_SIZE(i) > 1
+                     ? CVT_BLUES_SIZE(i) - (font->windows_compatibility ? 2
+                                                                        : 0)
+                     : 0));
   }
   *(bufp++) = num_used_styles;
   COPY_PREP(loop_cvt_b);
@@ -755,7 +813,10 @@ TA_table_build_prep(FT_Byte** prep,
     /* don't loop over artificial blue zones */
     *(bufp++) = (unsigned char)CVT_BLUE_SHOOTS_OFFSET(i);
     *(bufp++) = (unsigned char)(
-                  CVT_BLUES_SIZE(i) > 1 ? CVT_BLUES_SIZE(i) - 2 : 0);
+                  CVT_BLUES_SIZE(i) > 1
+                  ? CVT_BLUES_SIZE(i) - (font->windows_compatibility ? 2
+                                                                     : 0)
+                  : 0);
   }
   *(bufp++) = num_used_styles;
   COPY_PREP(loop_cvt_c);
@@ -812,6 +873,8 @@ TA_table_build_prep(FT_Byte** prep,
   COPY_PREP(set_smooth_or_strong_c);
   *(bufp++) = font->dw_cleartype_strong_stem_width ? 100 : 0;
   COPY_PREP(set_smooth_or_strong_d);
+  *(bufp++) = font->dw_cleartype_strong_stem_width ? 100 : 0;
+  COPY_PREP(set_smooth_or_strong_e);
 
   if (num_used_styles > 3)
   {
