@@ -1,7 +1,7 @@
 /* ttfautohint.c */
 
 /*
- * Copyright (C) 2011-2016 by Werner Lemberg.
+ * Copyright (C) 2011-2017 by Werner Lemberg.
  *
  * This file is part of the ttfautohint library, and may only be used,
  * modified, and distributed under the terms given in `COPYING'.  By
@@ -16,6 +16,7 @@
 /* This file needs FreeType 2.4.5 or newer. */
 
 #include <stdio.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -60,12 +61,18 @@ TTF_autohint(const char* options,
   FILE* out_file = NULL;
   FILE* control_file = NULL;
 
+  FILE* reference_file = NULL;
+  FT_Long reference_index = 0;
+  const char* reference_name = NULL;
+
   const char* in_buf = NULL;
   size_t in_len = 0;
   char** out_bufp = NULL;
   size_t* out_lenp = NULL;
   const char* control_buf = NULL;
   size_t control_len = 0;
+  const char* reference_buf = NULL;
+  size_t reference_len = 0;
 
   const unsigned char** error_stringp = NULL;
 
@@ -96,15 +103,18 @@ TTF_autohint(const char* options,
   FT_Bool adjust_subglyphs = 0;
   FT_Bool hint_composites = 0;
   FT_Bool symbol = 0;
+  FT_Bool fallback_scaling = 0;
 
   const char* fallback_script_string = NULL;
   const char* default_script_string = NULL;
+
   TA_Style fallback_style = TA_STYLE_NONE_DFLT;
   TA_Script default_script = TA_SCRIPT_LATN;
 
   FT_Bool dehint = 0;
   FT_Bool debug = 0;
   FT_Bool TTFA_info = 0;
+  unsigned long long epoch = ULLONG_MAX;
 
   const char* op;
 
@@ -146,12 +156,6 @@ TTF_autohint(const char* options,
     /* handle options -- don't forget to update parameter dump below! */
     if (COMPARE("adjust-subglyphs"))
       adjust_subglyphs = (FT_Bool)va_arg(ap, FT_Int);
-    else if (COMPARE("debug"))
-      debug = (FT_Bool)va_arg(ap, FT_Int);
-    else if (COMPARE("default-script"))
-      default_script_string = va_arg(ap, const char*);
-    else if (COMPARE("dehint"))
-      dehint = (FT_Bool)va_arg(ap, FT_Int);
     else if (COMPARE("control-buffer"))
     {
       control_file = NULL;
@@ -168,14 +172,24 @@ TTF_autohint(const char* options,
       control_buf = NULL;
       control_len = 0;
     }
+    else if (COMPARE("debug"))
+      debug = (FT_Bool)va_arg(ap, FT_Int);
+    else if (COMPARE("default-script"))
+      default_script_string = va_arg(ap, const char*);
+    else if (COMPARE("dehint"))
+      dehint = (FT_Bool)va_arg(ap, FT_Int);
     else if (COMPARE("dw-cleartype-strong-stem-width"))
       dw_cleartype_strong_stem_width = (FT_Bool)va_arg(ap, FT_Int);
+    else if (COMPARE("epoch"))
+      epoch = (unsigned long long)va_arg(ap, unsigned long long);
     else if (COMPARE("error-callback"))
       err = va_arg(ap, TA_Error_Func);
     else if (COMPARE("error-callback-data"))
       err_data = va_arg(ap, void*);
     else if (COMPARE("error-string"))
       error_stringp = va_arg(ap, const unsigned char**);
+    else if (COMPARE("fallback-scaling"))
+      fallback_scaling = (FT_Bool)va_arg(ap, FT_Int);
     else if (COMPARE("fallback-script"))
       fallback_script_string = va_arg(ap, const char*);
     else if (COMPARE("fallback-stem-width"))
@@ -240,6 +254,26 @@ TTF_autohint(const char* options,
       progress = va_arg(ap, TA_Progress_Func);
     else if (COMPARE("progress-callback-data"))
       progress_data = va_arg(ap, void*);
+    else if (COMPARE("reference-buffer"))
+    {
+      reference_file = NULL;
+      reference_buf = va_arg(ap, const char*);
+    }
+    else if (COMPARE("reference-buffer-len"))
+    {
+      reference_file = NULL;
+      reference_len = va_arg(ap, size_t);
+    }
+    else if (COMPARE("reference-file"))
+    {
+      reference_file = va_arg(ap, FILE*);
+      reference_buf = NULL;
+      reference_len = 0;
+    }
+    else if (COMPARE("reference-index"))
+      reference_index = (FT_Long)va_arg(ap, FT_UInt);
+    else if (COMPARE("reference-name"))
+      reference_name = va_arg(ap, const char*);
     else if (COMPARE("symbol"))
       symbol = (FT_Bool)va_arg(ap, FT_Int);
     else if (COMPARE("TTFA-info"))
@@ -377,6 +411,9 @@ TTF_autohint(const char* options,
     }
   }
 
+  font->reference_index = reference_index;
+  font->reference_name = reference_name;
+
   font->hinting_range_min = (FT_UInt)hinting_range_min;
   font->hinting_range_max = (FT_UInt)hinting_range_max;
   font->hinting_limit = (FT_UInt)hinting_limit;
@@ -393,6 +430,7 @@ TTF_autohint(const char* options,
   font->adjust_subglyphs = adjust_subglyphs;
   font->hint_composites = hint_composites;
   font->fallback_style = fallback_style;
+  font->fallback_scaling = fallback_scaling;
   font->default_script = default_script;
   font->symbol = symbol;
 
@@ -406,6 +444,7 @@ No_check:
   font->debug = debug;
   font->dehint = dehint;
   font->TTFA_info = TTFA_info;
+  font->epoch = epoch;
 
   font->gasp_idx = MISSING;
 
@@ -413,7 +452,7 @@ No_check:
 
   if (in_file)
   {
-    error = TA_font_file_read(font, in_file);
+    error = TA_font_file_read(in_file, &font->in_buf, &font->in_len);
     if (error)
       goto Err;
   }
@@ -441,6 +480,26 @@ No_check:
     font->control_len = control_len;
   }
 
+  if (reference_file)
+  {
+    error = TA_font_file_read(reference_file,
+                              &font->reference_buf,
+                              &font->reference_len);
+    if (error)
+      goto Err;
+  }
+  else if (reference_buf)
+  {
+    /* a valid TTF can never be that small */
+    if (reference_len < 100)
+    {
+      error = TA_Err_Invalid_Font_Type + 0x300;
+      goto Err1;
+    }
+    font->reference_buf = (FT_Byte*)reference_buf;
+    font->reference_len = reference_len;
+  }
+
   error = TA_font_init(font);
   if (error)
     goto Err;
@@ -468,7 +527,7 @@ No_check:
     /* assure that the font hasn't been already processed by ttfautohint; */
     /* another, more thorough check is done in TA_glyph_parse_simple */
     idx = FT_Get_Name_Index(sfnt->face, (FT_String*)TTFAUTOHINT_GLYPH);
-    if (idx)
+    if (idx && !dehint)
     {
       error = TA_Err_Already_Processed;
       goto Err;
@@ -632,8 +691,7 @@ No_check:
     if (error)
       goto Err;
 
-    if (font->loader)
-      ta_loader_done(font);
+    ta_loader_done(font);
   }
 
   for (i = 0; i < font->num_sfnts; i++)
@@ -697,15 +755,24 @@ No_check:
 Err:
   TA_control_free(font->control);
   TA_control_free_tree(font);
-  TA_font_unload(font, in_buf, out_bufp, control_buf);
+  TA_font_unload(font, in_buf, out_bufp, control_buf, reference_buf);
 
 Err1:
-  if (!error_string)
-    error_string = (char*)TA_get_error_message(error);
+  {
+    FT_Error e = error;
 
-  /* this must be a static value */
-  if (error_stringp)
-    *error_stringp = (const unsigned char*)TA_get_error_message(error);
+
+    /* use standard FreeType error strings for reference file errors */
+    if (error >= 0x300 && error < 0x400)
+      e -= 0x300;
+
+    if (!error_string)
+      error_string = (char*)TA_get_error_message(e);
+
+    /* this must be a static value */
+    if (error_stringp)
+      *error_stringp = (const unsigned char*)TA_get_error_message(e);
+  }
 
   if (err)
     err(error,
